-- HZ Escorts Enhanced
-- Copyright (C) 2014 - 2021 Ben S. (@HunterZ0)
--
-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009 - 2019 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

local Astar = require "engine.Astar"

-- AI for the escort quest
-- the NPC will run toward the portal, if hostiles are in sight he attacks them
newAI("escort_quest", function(self)
	self.escort_path = rng.percent(10) -- repurposed as rest AI chat flag
	if self.ai_state.tactic_escort_rest then
		self.ai_state.tactic_escort_rest = self.ai_state.tactic_escort_rest - 1
		-- Rest
		if self.ai_state.tactic_escort_rest > 0 then return self:runAI("escort_rest")
		-- Cooldown
		elseif self.ai_state.tactic_escort_rest < -20 then self.ai_state.tactic_escort_rest = nil
		end
	end

	if self:runAI("target_simple") then
		-- One in "talent_in" chance of using a talent
		if rng.chance(self.ai_state.talent_in or 6) and self:reactionToward(self.ai_target.actor) < 0 then
			self:runAI("dumb_talented")
		end
		if not self.energy.used then
			local dist = self.ai_target.actor and core.fov.distance(self.x, self.y, self.ai_target.actor.x, self.ai_target.actor.y)
			if self:reactionToward(self.ai_target.actor) < 0 and dist <= 10 and self:hasLOS(self.ai_target.actor.x, self.ai_target.actor.y, "block_move") and not self:hasLOS(self.escort_target.x, self.escort_target.y, "block_move") then
				self:runAI("flee_dmap")
				if not self.ai_state.fleeing_msg then
					self.ai_state.fleeing_msg = true
					local enemy = self.ai_target.actor
					local dir = game.level.map:compassDirection(enemy.x-self.x, enemy.y-self.y)
					self:doEmote(_t"Help!"..(dir and (" %s to the %s!"):tformat(self:canSee(enemy) and string.capitalize(enemy:getName()) or _t"Something", dir) or ""))
				end
			else
				self:runAI("move_escort")
				self.ai_state.fleeing_msg = nil
			end
		end
		return true
	end

	if not self.energy.used then
		self:runAI("move_escort")
		return true
	end
end)

newAI("move_escort", function(self)
	if not self.escort_target then return true end

	local px, py = game.player.x, game.player.y
	local tx, ty = self.escort_target.x, self.escort_target.y
	local a = Astar.new(game.level.map, self)
	local path = nil
	local dist = core.fov.distance(tx, ty, self.x, self.y)
	local doChat = rng.percent(10)
	local doRest = rng.percent(35)
	local chatted = nil
	-- action categories:
	-- nil => TBD
	--  1  => rest
	--  2  => path to portal
	--  3  => path to player
	--  4  => wander due to blocked
	--  5  => wander due to lost
	local action = nil

	-- repurpose this as a way to pass doChat to rest AI
	self.escort_path = doChat

	-- if portal is nearby
	if dist < 8 then
		if doRest then
			action = 1
		else
			path = a:calc(self.x, self.y, tx, ty)
			if path and path[1] then action = 2 end
		end
	end

	-- if portal is far, or pathing failed
	if not action then
		-- don't randomly stop, as we want to reach player ASAP
		-- path towards player
		path = a:calc(self.x, self.y, px, py)
		if path and path[1] then action = 3 end
	end

	-- if we have a path, but player is in the way
	if path and path[1] and path[1].x == px and path[1].y == py then
		if doRest then
			action = 1
		else
			action = 4
		end
	end

	-- if no action chosen yet
	if not action then
		if doRest then
			action = 1
		else
			action = 5
		end
	end

	-- if randomly decided to chat
	if doChat then
		-- emote to player based on chosen action
		-- reserve an even chance of doing a random emote instead
		if 1 == action then
			-- chat handled in rest AI
			-- don't assume chat happens, as worst result is random chat happens alonside rest chat
		elseif 2 == action then
			if rng.percent(100 / 3) then
				self:doEmote(_t"The portal is nearby!")
				chatted = true
			elseif rng.percent(100 / 2) then
				self:doEmote(_t"I am making a run for the portal!")
				chatted = true
			end
		elseif 3 == action then
			if rng.percent(100 / 3) then
				self:doEmote(_t"I'm following you.")
				chatted = true
			elseif rng.percent(100 / 2) then
				self:doEmote(_t"Please help me reach the portal.")
				chatted = true
			end
		elseif 4 == action then
			if rng.percent(100 / 5) then
				self:doEmote(_t"We must hurry!")
				chatted = true
			elseif rng.percent(100 / 4) then
				self:doEmote(_t"Why have you stopped?")
				chatted = true
			elseif rng.percent(100 / 3) then
				self:doEmote(_t"Let's go!")
				chatted = true
			elseif rng.percent(100 / 2) then
				self:doEmote(_t"I must reach the portal!")
				chatted = true
			end
		else
			if rng.percent(100 / 3) then
				self:doEmote(_t"I seem to be lost.")
				chatted = true
			elseif rng.percent(100 / 2) then
				self:doEmote(_t"I have lost my way.")
				chatted = true
			end
		end

		-- if haven't chatted
		if not chatted then
			-- emote random stuff to the player
			if rng.percent(100 / 3) then
				self:doEmote(_t"I am glad I found you.")
				chatted = true
			elseif rng.percent(100 / 2) then
				self:doEmote(_t"I am grateful for your aid.")
				chatted = true
			else -- 100 percent
				local dir = game.level.map:compassDirection(tx - self.x, ty - self.y)
				if dir then
					local distName = _t"still far away"
					if dist < 8 then
						distName = _t"very close"
					elseif dist < 16 then
						distName = _t"close"
					end
					self:doEmote(("The portal is %s, to the %s."):tformat(distName, dir))
					chatted = true
				else
					self:doEmote(_t"I am uncertain of the portal's direction.")
					chatted = true
				end
			end
		end
	end -- doChat

	-- take appropriate action
	if 1 == action then
		-- rest
		return self:runAI("escort_rest")
	end
	self.escort_path = nil
	if 2 == action or 3 == action then
		-- move along path
		return self:move(path[1].x, path[1].y)
	end
	-- wander
	return self:runAI("move_simple")
end)

newAI("escort_rest", function(self)
	local healthGainPercent = 0.05
	local doChat = self.escort_path
	self.escort_path = nil

	-- if wounded
	if self.life < self.max_life then
		-- report status to player
		if doChat then
			if rng.percent(100 / 3) then
				self:doEmote(_t"I am taking a quick break.")
			elseif rng.percent(100 / 2) then
				self:doEmote(_t"I am tending to my wounds.")
			else
				self:doEmote(_t"I am feeling a little better.")
			end
		end

		-- gain some health
		self.life = self.life + (self.max_life * healthGainPercent)
		if self.life > self.max_life then
			-- clamp to max
			self.life = self.max_life
		end
	elseif doChat then
		if rng.percent(100 / 3) then
			self:doEmote(_t"I am at full strength.")
		elseif rng.percent(100 / 2) then
			self:doEmote(_t"I am ready to fight!")
		else
			self:doEmote(_t"I am good to go.")
		end
	end

	-- pass turn
	self:useEnergy()
	return true
end)

